{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "fb921616",
   "metadata": {},
   "source": [
    "### CGO 12-3: Colorization\n",
    "\n",
    "> **Let there be Color!: Joint End-to-end Learning of Global and Local Image Priors for Automatic Image Colorization with Simultaneous Classification**\n",
    "> Satoshi Iizuka\\*, Edgar Simo-Serra\\*, Hiroshi Ishikawa (* equal contribution)\n",
    "> ACM Transactions on Graphics (SIGGRAPH), 2016\n",
    "\n",
    "#### About colorization\n",
    "\n",
    "* [Website](http://iizuka.cs.tsukuba.ac.jp/projects/colorization/)\n",
    "* [Paper](https://esslab.jp/publications/IizukaSIGGRAPH2016.pdf)\n",
    "\n",
    "#### Objectives\n",
    "\n",
    "1. Test baseline\n",
    "1. Finish implementing global and local network\n",
    "1. Explain why the global feature network helps (hint: spatial resolution)\n",
    "1. Add classification loss (hint: it is possible to modify forward() to return multiple values. Set the classification loss to 1/300 of the reconstruction loss)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0b63de69",
   "metadata": {},
   "source": [
    "PUT ANSWER TO WHY GLOBAL FEATURES HELP HERE"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d83d09f9",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "import torch.nn as nn\n",
    "import torch.nn.functional as F"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f1c89fbf",
   "metadata": {},
   "source": [
    "<img src='https://esslab.jp/~ess/images/teaching/cgo/11_colorization_model.png' alt='Colorization network' style='max-width:800px;'/>\n",
    "\n",
    "####  Regarding the model\n",
    "\n",
    "* Conv: convolutional layer\n",
    "* FC: fully connected layer\n",
    "* Fusion: fusion layer\n",
    "* Upsample: upscaling layer\n",
    "* Apply TanH to ouptut\n",
    "\n",
    "#### Improvements\n",
    "\n",
    "1. Reduced filters from 512 to 256\n",
    "1. Last layer restores to original resolution"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1e6b9c09",
   "metadata": {},
   "outputs": [],
   "source": [
    "class ConvLayer(nn.Module):\n",
    "    def __init__(self, fan_in, fan_out, kernel=3, stride=1, padding=1, upscale=None):\n",
    "        super(ConvLayer, self).__init__()\n",
    "        self.conv = nn.Conv2d( fan_in, fan_out, kernel_size=kernel, stride=stride, padding=padding, bias=False )\n",
    "        self.bn   = nn.BatchNorm2d( fan_out )\n",
    "        self.upscale = upscale\n",
    "    def forward(self, x):\n",
    "        if self.upscale != None:\n",
    "            x = F.interpolate( x, scale_factor=self.upscale, mode='bilinear', align_corners=False)\n",
    "        return F.relu( self.bn( self.conv(x) ) )\n",
    "    ''\n",
    "class FCLayer(nn.Module):\n",
    "    def __init__(self, fan_in, fan_out ):\n",
    "        super(ConvLayer, self).__init__()\n",
    "        self.fc = nn.Linear( fan_in, fan_out )\n",
    "        self.bn = nn.BatchNorm( fan_out )\n",
    "    def forward(self, x):\n",
    "        return F.relu( self.bn( self.fc(x) ) )\n",
    "\n",
    "class FusionLayer(nn.Module):\n",
    "    def __init__(self, fan_in, fan_out, inner_size ):\n",
    "        super(FusionLayer, self).__init__()\n",
    "        self.conv = ConvLayer( fan_in, fan_out, kernel=1, padding=0 )\n",
    "        self.inner_size = inner_size\n",
    "    def forward(self, l, g):\n",
    "        x = torch.cat( [l, g.repeat(1,1,self.inner_size,self.inner_size)], dim=1 )\n",
    "        return self.conv(x)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "741d8bcf",
   "metadata": {},
   "source": [
    "#### Model Class"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e4b03f55",
   "metadata": {},
   "outputs": [],
   "source": [
    "class Net(nn.Module):\n",
    "    def __init__(self, use_global=True, use_tanh=True):\n",
    "        super(Net, self).__init__()\n",
    "        self.use_global = use_global\n",
    "        self.use_tanh = use_tanh\n",
    "\n",
    "        self.lowlevel = nn.Sequential(\n",
    "            ConvLayer(   1,  64, stride=2 ), # 64\n",
    "            ConvLayer(  64, 128 ),\n",
    "            ConvLayer( 128, 128, stride=2 ), # 32\n",
    "            ConvLayer( 128, 256 ),\n",
    "            ConvLayer( 256, 256, stride=2 ), # 16\n",
    "            ConvLayer( 256, 256 )\n",
    "        )\n",
    "\n",
    "        self.midlevel = nn.Sequential(\n",
    "            ConvLayer( 256, 256 ),\n",
    "            ConvLayer( 256, 256 )\n",
    "        )\n",
    "\n",
    "        if self.use_global:\n",
    "            self.globalfeat = nn.Sequential(\n",
    "                ConvLayer( 256, 256, stride=2 ), # 8\n",
    "                ConvLayer( 256, 256 ),\n",
    "                ConvLayer( 256, 256, stride=2 ), # 4\n",
    "                ConvLayer( 256, 256 ),\n",
    "                ConvLayer( 256, 512, kernel=4, padding=0 ), # 1\n",
    "                ConvLayer( 512, 256, kernel=1, padding=0 )\n",
    "            )\n",
    "\n",
    "            self.fusion = FusionLayer( 512, 256, 16 )\n",
    "\n",
    "            self.classifier = nn.Sequential(\n",
    "                ConvLayer( 256, 128, kernel=1, padding=0 ),\n",
    "                nn.Conv2d( 128,   6, kernel_size=1, padding=0 )\n",
    "            )\n",
    "\n",
    "        self.colorization = nn.Sequential(\n",
    "            ConvLayer( 256, 128 ),\n",
    "            ConvLayer( 128,  64, upscale=2. ),\n",
    "            ConvLayer(  64,  64 ),\n",
    "            ConvLayer(  64,  32, upscale=2. ),\n",
    "            ConvLayer(  32,  16, upscale=2. ),\n",
    "            nn.Conv2d(  16,   2, kernel_size=3, padding=1 )\n",
    "        )\n",
    "\n",
    "    def forward( self, x ):\n",
    "        l = self.lowlevel( x )\n",
    "        m = self.midlevel( l )\n",
    "        c = self.colorization( m )\n",
    "        if self.use_tanh:\n",
    "            c = torch.tanh( c )\n",
    "        return torch.cat( [x, c], dim=1 )\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "26ee58de",
   "metadata": {},
   "source": [
    "#### Preparing data\n",
    "\n",
    "Download from [here](https://esslab.jp/~ess/data/tinyplaces.tar.gz) and unzip into `../data`.\n",
    "\n",
    "1. 6 classes from places (bamboo_forest, coast, corn_field, sea_cliff, tree_farm, volcano)\n",
    "1. class labels available\n",
    "1. Converted to CIE Lab colorspace"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "dfb6bdcb",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torchvision\n",
    "from skimage import color\n",
    "import numpy as np\n",
    "from PIL import Image\n",
    "import random\n",
    "def lab_loader(path):\n",
    "    # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)\n",
    "    with open(path, 'rb') as f:\n",
    "        img = Image.open(f)\n",
    "        rgb = img.convert('RGB')\n",
    "        lab = color.rgb2lab(rgb)\n",
    "        return lab.astype(np.float32)\n",
    "transform = torchvision.transforms.Compose([\n",
    "    torchvision.transforms.ToTensor(),\n",
    "    torchvision.transforms.Normalize((50,0,0,), (50,128,128,))\n",
    "])\n",
    "data = torchvision.datasets.ImageFolder('../data/tinyplaces/', transform=transform, loader=lab_loader)\n",
    "\n",
    "idx       = list(range(len(data)))\n",
    "random.shuffle(idx)\n",
    "idx       = idx[0:10000]\n",
    "limit     = int(0.1*len(idx))\n",
    "test_idx  = idx[:limit]\n",
    "train_idx = idx[limit:]\n",
    "train_sampler = torch.utils.data.sampler.SubsetRandomSampler( train_idx )\n",
    "train_loader = torch.utils.data.DataLoader( data, batch_size=16, sampler=train_sampler )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8df6337b",
   "metadata": {},
   "source": [
    "#### Create model and optimization approach"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c398a444",
   "metadata": {},
   "outputs": [],
   "source": [
    "device = torch.device(\"cuda\" if (torch.cuda.device_count()>0) else \"cpu\")\n",
    "model = Net().to(device)\n",
    "optimizer = torch.optim.Adadelta( model.parameters() )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d72a6ec8",
   "metadata": {},
   "source": [
    "#### Training loop"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "673aa057",
   "metadata": {},
   "outputs": [],
   "source": [
    "loss_log = []"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d4504698",
   "metadata": {},
   "outputs": [],
   "source": [
    "from tqdm import tqdm\n",
    "model.train()\n",
    "t = tqdm(train_loader)\n",
    "for x,y in t:\n",
    "    optimizer.zero_grad()\n",
    "    x, y = x.to(device), y.to(device)\n",
    "    L  = x[:,0:1,:,:].detach()\n",
    "    ab = x[:,1:3,:,:].detach()\n",
    "\n",
    "    pred = model(L)\n",
    "    loss = torch.pow( pred-x, 2 ).mean()\n",
    "    t.set_postfix( loss=loss.item() )\n",
    "    loss_log.append( loss.item() )\n",
    "    loss.backward()\n",
    "    optimizer.step()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7ef28eb0",
   "metadata": {},
   "outputs": [],
   "source": [
    "%matplotlib inline\n",
    "import matplotlib.pyplot as plt\n",
    "plt.plot( loss_log, label='Loss' )\n",
    "plt.legend( frameon=False )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "605c11b0",
   "metadata": {},
   "source": [
    "#### Results"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "25180abe",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torchvision\n",
    "import matplotlib.pyplot as plt\n",
    "\n",
    "def tolab( x ):\n",
    "    x = x[0].clone()\n",
    "    x[0,:,:]  = 50 + 50*x[0,:,:]\n",
    "    x[1,:,:] *= 128\n",
    "    x[2,:,:] *= 128\n",
    "    return torch.Tensor( color.lab2rgb( x.permute(1,2,0) )).permute(2,0,1).unsqueeze(0)\n",
    "\n",
    "model.eval()\n",
    "N = 10\n",
    "I = torch.zeros( N*2, 3, 128, 128 )\n",
    "with torch.no_grad():\n",
    "    for i,idx in enumerate(test_idx[:N]):\n",
    "        x  = data[idx][0].unsqueeze(0)\n",
    "        L  = x[:,0:1,:,:]\n",
    "        ab = x[:,1:3,:,:]\n",
    "\n",
    "        pred = model(L.to(device)).cpu()\n",
    "        I[i,:,:,:]   = tolab( x )\n",
    "        I[N+i,:,:,:] = tolab( pred )\n",
    "\n",
    "I = torchvision.utils.make_grid( I, nrow=5 )\n",
    "plt.imshow( I.permute(1, 2, 0) )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bec65d20",
   "metadata": {},
   "source": [
    "#### Spatial resolution"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3bb99a4c",
   "metadata": {},
   "outputs": [],
   "source": [
    "I = torch.zeros(1,1,128,128)\n",
    "I[0,0,63,63] = 1\n",
    "m = Net( use_global=True, use_tanh=False )\n",
    "for name, p in m.named_parameters():\n",
    "    if 'weight' in name:\n",
    "        nn.init.constant_(p,1)\n",
    "    elif 'bias' in name:\n",
    "        nn.init.constant_(p,0)\n",
    "m.double()\n",
    "m.eval()\n",
    "with torch.no_grad():\n",
    "    I = m( I.double() )\n",
    "plt.imshow(I[0,1,:,:] / I[0,1,:,:].max() )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c5bb438b",
   "metadata": {},
   "source": [
    "#### Display model\n",
    "\n",
    "https://github.com/waleedka/hiddenlayer\n",
    "\n",
    "###### Install\n",
    "\n",
    "````\n",
    "conda install graphviz python-graphviz\n",
    "pip install hiddenlayer\n",
    "````\n",
    "\n",
    "```python\n",
    "import hiddenlayer as hl\n",
    "hl.build_graph( model.cpu(), torch.zeros([1, 1, 128, 128]) )\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "49f8ab1a",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "jupytext": {
   "cell_metadata_filter": "-all",
   "main_language": "python",
   "notebook_metadata_filter": "-all"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}